#!/usr/bin/env python

# Copyright 2011 Ian Goldberg
# Copyright 2016 George Danezis (UCL InfoSec Group)
#
# This file is part of Sphinx.
# 
# Sphinx is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU Lesser General Public
# License as published by the Free Software Foundation.
# 
# Sphinx is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public
# License along with Sphinx.  If not, see
# <http://www.gnu.org/licenses/>.

from os import urandom
from struct import unpack
from binascii import hexlify

# Python 2/3 compatibility
from builtins import bytes

from . import SphinxException

# Core Process function -- devoid of any chrome
def sphinx_process(params, secret, header, delta):
    """ The heart of a Sphinx server, that processes incoming messages.
    It takes a set of parameters, the secret of the server,
    and an incoming message header and body.
        
    """
    p = params
    group = p.group
    alpha, beta, gamma = header

    # Check that alpha is in the group
    if not group.in_group(alpha):
        raise SphinxException("Alpha not in Group.")

    # Compute the shared secret
    s = group.expon(alpha, secret)
    aes_s = p.get_aes_key(s)
    
    assert len(beta) == p.max_len - 32
    # print("B: \n%s" % hexlify(beta))
    if gamma != p.mu(p.hmu(aes_s), beta):
        raise SphinxException("MAC mismatch.")

    beta_pad = beta + (b"\x00" * (2 * p.max_len)) 
    B = p.xor_rho(p.hrho(aes_s), beta_pad)

    length = B[0]
    routing = B[1:1+length]
    rest = B[1+length:]

    tag = p.htau(aes_s)
    b = p.hb(alpha, aes_s)
    alpha = group.expon(alpha, b)
    gamma = rest[:p.k]
    beta = rest[p.k:p.k+(p.max_len - 32)]
    delta = p.pii(p.hpi(aes_s), delta)

    ret = (tag, routing, ((alpha, beta, gamma), delta))
    return ret




# -*- coding: utf-8 -*-
"""The ``sphinxmix`` package implements the Sphinx mix packet format core cryptographic functions.

The paper describing sphinx may be found here:

    * George Danezis and Ian Goldberg. Sphinx: A Compact and Provably Secure Mix Format. IEEE Symposium on Security and Privacy 2009. [`link <http://www.cypherpunks.ca/~iang/pubs/Sphinx_Oakland09.pdf>`_]


All the ``sphinxmix`` cryptography is encapsulated and within a ``SphinxParams`` object that 
is used by all subsequent functions. To make ``sphinxmix`` use different cryptographic 
primitives simply extend this class, or re-implement it. The default cryptographic primitives 
are ``NIST/SEGS-p224`` curves, ``AES`` and ``SHA256``.

Sending Sphinx messages
-----------------------

To package or process sphinx messages create a new ``SphinxParams`` object:

    >>> # Instantiate a the crypto parameters for Sphinx.
    >>> from sphinxmix.SphinxParams import SphinxParams
    >>> params = SphinxParams()

The ``sphinxmix`` package requires some rudimentary Public Key Information: mix nodes need
an identifier created by ``Nenc`` and the PKI consists of a dictionary mapping node names
to ``pki_entry`` records. Those include secret keys (derived using ``gensecret``) and public 
keys (derived using ``expon``).

    >>> # The minimal PKI involves names of nodes and keys
    >>> from sphinxmix.SphinxClient import pki_entry, Nenc
    >>> pkiPriv = {}
    >>> pkiPub = {}
    >>> for i in range(10):
    ...     nid = i
    ...     x = params.group.gensecret()
    ...     y = params.group.expon(params.group.g, x)
    ...     pkiPriv[nid] = pki_entry(nid, x, y)
    ...     pkiPub[nid] = pki_entry(nid, None, y)

A client may package a message using the Sphinx format to relay over a number of mix servers. 
The function ``rand_subset`` may be used to select a random number of node identifiers; the function
``create_forward_message`` can then be used to package the message, ready to be sent to the 
first mix. Note both destination and message need to be ``bytes``.

    >>> # The simplest path selection algorithm and message packaging
    >>> from sphinxmix.SphinxClient import rand_subset, \\
    ...                                    create_forward_message
    >>> use_nodes = rand_subset(pkiPub.keys(), 5)
    >>> nodes_routing = list(map(Nenc, use_nodes))
    >>> keys_nodes = [pkiPub[n].y for n in use_nodes]
    >>> dest = b"bob"
    >>> message = b"this is a test"
    >>> header, delta = create_forward_message(params, nodes_routing, \\
    ...     keys_nodes, dest, message)

The client may specify any information in the ``nodes_routing`` list, that will
be passed to intermediate mixes. At a minimum this should include information about 
the next mix.

Processing Sphinx messages at a mix
-----------------------------------

The heart of a Sphinx mix server is the ``sphinx_process`` function, that takes the server
secret and decodes incoming messages. In this example the message encode above, is decoded
by the sequence of mixes.

    >>> # Process message by the sequence of mixes
    >>> from sphinxmix.SphinxClient import PFdecode, Relay_flag, Dest_flag, Surb_flag, receive_forward
    >>> from sphinxmix.SphinxNode import sphinx_process
    >>> x = pkiPriv[use_nodes[0]].x
    >>> while True:
    ...     ret = sphinx_process(params, x, header, delta)
    ...     (tag, info, (header, delta)) = ret
    ...     routing = PFdecode(params, info)
    ...     if routing[0] == Relay_flag:
    ...         flag, addr = routing
    ...         x = pkiPriv[addr].x 
    ...     elif routing[0] == Dest_flag:
    ...         assert receive_forward(params, delta) == [dest, message]
    ...         break

It is the responsibility of a mix to record ``tags`` of messages to prevent 
replay attacks. The ``PFdecode`` function may be used to recover routing Information
including the next mix, or any other user specified information.

Single use reply Blocks
-----------------------

A facility provided by Sphinx is the creation and use of Single Use Reply Blocks (SURB)
to route messages back to an anonymous receipient. First a receiver needs to create
a SURB using ``create_surb`` and passes on the ``nymtuple`` structure to the sender, and
storing ``surbkeytuple`` keyed by the identifier ``surbid``:

    >>> from sphinxmix.SphinxClient import create_surb, package_surb
    >>> surbid, surbkeytuple, nymtuple = create_surb(params, nodes_routing, keys_nodes, b"myself")

Using the ``nymtuple`` a sender can package a message to be sent through the network, 
starting at the ``nymtuple[0]`` router:

    >>> message = b"This is a reply"
    >>> header, delta = package_surb(params, nymtuple, message)

The network processes the SURB as any other message, until it is received 
by the last mix in the path:

    >>> x = pkiPriv[use_nodes[0]].x
    >>> while True:
    ...    ret = sphinx_process(params, x, header, delta)
    ...    (tag, B, (header, delta)) = ret
    ...    routing = PFdecode(params, B)
    ...
    ...    if routing[0] == Relay_flag:
    ...        flag, addr = routing
    ...        x = pkiPriv[addr].x 
    ...    elif routing[0] == Surb_flag:
    ...        flag, dest, myid = routing
    ...        break

The final mix server must sent the ``myid`` and ``delta`` to the destination
``dest``, where it may be decoded using the ``surbkeytuple``.

    >>> from sphinxmix.SphinxClient import receive_surb
    >>> received = receive_surb(params, surbkeytuple, delta)
    >>> assert received == message

Embedding arbitrary information for mixes
-----------------------------------------

A sender may embed arbitrary information to mix nodes, as demonstrated 
by embedding ``b'info'`` to each mix, and ``b'final_info'`` to the final 
mix:

    >>> use_nodes = rand_subset(pkiPub.keys(), 5)
    >>> nodes_routing = [Nenc((n, b'info')) for n in use_nodes]
    >>> keys_nodes = [pkiPub[n].y for n in use_nodes]
    >>> dest = (b"bob", b"final_info")
    >>> message = b"this is a test"
    >>> header, delta = create_forward_message(params, nodes_routing, \\
    ...     keys_nodes, dest, message)

Mixes decode the arbitrary structure passed by the clients, and can interpret
it to implement more complex mixing strategies:

    >>> x = pkiPriv[use_nodes[0]].x
    >>> while True:
    ...     ret = sphinx_process(params, x, header, delta)
    ...     (tag, info, (header, delta)) = ret
    ...     routing = PFdecode(params, info)
    ...     if routing[0] == Relay_flag:
    ...         flag, (addr, additional_info) = routing
    ...         assert additional_info == b'info'
    ...         x = pkiPriv[addr].x 
    ...     elif routing[0] == Dest_flag:
    ...         [[dest, additional_info], msg] = receive_forward(params, delta)
    ...         assert additional_info == b'final_info'
    ...         assert dest == b'bob'
    ...         break


Packaging mix messages to byte strings:
---------------------------------------

The `sphinxmix` package provides functions `pack_message` and `unpack_message` to 
serialize and deserialize mix messages using `msgpack`. Some meta-data about the 
parameter length are passed along the message any may be used to select an appropriate
parameter environment for the decoding of the message.

    >>> from sphinxmix.SphinxClient import pack_message, unpack_message
    >>> bin_message = pack_message(params, (header, delta))
    >>> param_dict = { (params.max_len, params.m):params }
    >>> px, (header1, delta1) = unpack_message(param_dict, bin_message)


"""

VERSION = "0.0.6"

class SphinxException(Exception):
    pass


#!/usr/bin/env python

# Copyright 2011 Ian Goldberg
# Copyright 2016 George Danezis (UCL InfoSec Group)
#
# This file is part of Sphinx.
# 
# Sphinx is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU Lesser General Public
# License as published by the Free Software Foundation.
# 
# Sphinx is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public
# License along with Sphinx.  If not, see
# <http://www.gnu.org/licenses/>.


from os import urandom
from collections import namedtuple
from struct import pack
from binascii import hexlify
from petlib.pack import encode, decode

# Python 2/3 compatibility
from builtins import bytes

from .SphinxParams import SphinxParams
from . import SphinxException


# FLAGS

#: Routing flag indicating message is to be relayed.
Relay_flag = "\xF0"

#: Routing flag indicating message is to be delivered.
Dest_flag = "\xF1"

#: Routing flag indicating surb reply is to be delivered.
Surb_flag = "\xF2"

# Padding/unpadding of message bodies: a 0 bit, followed by as many 1
# bits as it takes to fill it up

header_record = namedtuple("header_record", ["alpha", "s", "b", "aes"])

#: A helper named tuple to store PKI information.
pki_entry = namedtuple("pki_entry", ["id", "x", "y"])


def pad_body(msgtotalsize, body):
    """ Unpad the Sphinx message body."""
    body = body + b"\x7f"
    body = body + (b"\xff" * (msgtotalsize - len(body)))

    if (msgtotalsize - len(body)) < 0:
        raise SphinxException("Insufficient space for body") 

    return body

def unpad_body(body):
    """ Pad a Sphinx message body. """
    body = bytes(body)
    l = len(body) - 1
    x_marker = bytes(b"\x7f")[0]
    f_marker = bytes(b"\xff")[0]
    while body[l] == f_marker and l > 0:
        l -= 1
    
    if body[l] == x_marker:
        ret = body[:l]
    else:
        ret = b''
    
    return ret

# Prefix-free encoding/decoding of node names and destinations

# Sphinx nodes
def Nenc(idnum):
    """ The encoding of mix names. """
    return Route_pack((Relay_flag, idnum))

def Route_pack(info):
    return encode(info)

# Decode the prefix-free encoding.  Return the type, value, and the
# remainder of the input string
def PFdecode(param, packed):
    """ Decoder of prefix free encoder for commands received by mix or clients."""
    assert type(packed) is bytes
    return decode(packed)


def rand_subset(lst, nu):
    """Return a list of nu random elements of the given list (without
    replacement)."""

    # Randomize the order of the list by sorting on a random key
    nodeids = [(urandom(8),x) for x in lst]
    nodeids.sort(key=lambda x:x[0])

    # Return the first nu elements of the randomized list
    return list(map(lambda x:x[1], nodeids[:nu]))


def create_header(params, nodelist, keys, dest):
    """ Internal function, creating a Sphinx header, given parameters, a node list (path), 
    a pki mapping node names to keys, a destination, and a message identifier.""" 

    node_meta = [pack("b", len(n)) + n for n in nodelist]

    p = params
    nu = len(nodelist)
    max_len = p.max_len
    
    group = p.group
    x = group.gensecret()

    blind_factor = x
    asbtuples = []
    
    for k in keys:
        alpha = group.expon(group.g, blind_factor)
        s = group.expon(k, blind_factor)
        aes_s = p.get_aes_key(s)

        b = p.hb(alpha, aes_s)
        blind_factor = blind_factor.mod_mul(b, p.group.G.order())

        hr = header_record(alpha, s, b, aes_s)
        asbtuples.append(hr)

    # Compute the filler strings
    phi = b''
    min_len = (max_len - 32)
    for i in range(1,nu):

        plain = phi + (b"\x00" * (p.k + len(node_meta[i])))
        phi = p.xor_rho(p.hrho(asbtuples[i-1].aes), (b"\x00"*min_len)+plain)
        phi = phi[min_len:]

        min_len -= len(node_meta[i]) + p.k
        
    
    assert len(phi) == sum(map(len, node_meta[1:])) + (nu-1)*p.k

    # Compute the (beta, gamma) tuples
    # The os.urandom used to be a string of 0x00 bytes, but that's wrong
    
    final_routing = pack("b", len(dest)) + dest

    len_meta = sum(map(len, node_meta[1:]))
    random_pad_len = (max_len - 32) - len_meta - (nu-1)*p.k - len(final_routing)

    if random_pad_len < 0:
        raise SphinxException("Insufficient space routing info") 

    beta = final_routing + urandom(random_pad_len)
    beta = p.xor_rho(p.hrho(asbtuples[nu-1].aes), beta) + phi

    gamma = p.mu(p.hmu(asbtuples[nu-1].aes), beta)
    
    for i in range(nu-2, -1, -1):
        node_id = node_meta[i+1]

        plain_beta_len = (max_len - 32) - p.k - len(node_id)
        
        plain = node_id + gamma + beta[:plain_beta_len]
        
        beta = p.xor_rho(p.hrho(asbtuples[i].aes), plain)        
        gamma = p.mu(p.hmu(asbtuples[i].aes), beta)
        
    return (asbtuples[0].alpha, beta, gamma), \
        [x.aes for x in asbtuples]


def create_forward_message(params, nodelist, keys, dest, msg):
    """Creates a forward Sphix message, ready to be processed by a first mix. 

    It takes as parameters a node list of mix information, that will be provided to each mix, forming the path of the message;
    a list of public keys of all intermediate mixes; a destination and a message (byte arrays)."""

    p = params
    # pki = p.pki
    nu = len(nodelist)
    assert len(dest) < 128 and len(dest) > 0
    assert p.k + 1 + len(dest) + len(msg) < p.m

    # Compute the header and the secrets

    final = Route_pack((Dest_flag, ))
    header, secrets = create_header(params, nodelist, keys, final)

    body = pad_body(p.m, (b"\x00" * p.k) + encode((dest, msg)))

    # Compute the delta values
    delta = p.pi(p.hpi(secrets[nu-1]), body)
    for i in range(nu-2, -1, -1):
        delta = p.pi(p.hpi(secrets[i]), delta)

    return header, delta

def create_surb(params, nodelist, keys, dest):
    """Creates a Sphinx single use reply block (SURB) using a set of parameters;
    a sequence of mix identifiers; a pki mapping names of mixes to keys; and a final 
    destination.

    Returns:
        - A triplet (surbid, surbkeytuple, nymtuple). Where the surbid can be 
          used as an index to store the secrets surbkeytuple; nymtuple is the actual
          SURB that needs to be sent to the receiver.

    """
    p = params
    nu = len(nodelist)
    xid = urandom(p.k)

    # Compute the header and the secrets
    final = Route_pack((Surb_flag, dest, xid))
    header, secrets = create_header(params, nodelist, keys, final )

    ktilde = urandom(p.k)
    keytuple = [ktilde]
    keytuple.extend(map(p.hpi, secrets))
    return xid, keytuple, (nodelist[0], header, ktilde)

def package_surb(params, nymtuple, message):
    """Packages a message to be sent with a SURB. The message has to be bytes, 
    and the nymtuple is the structure returned by the create_surb call.

    Returns a header and a body to pass to the first mix.
    """
    n0, header0, ktilde = nymtuple
    body = params.pi(ktilde, pad_body(params.m, (b"\x00" * params.k) + message))
    return (header0, body)


def receive_forward(params, delta):
    """ Decodes the body of a forward message."""
    
    if delta[:params.k] != b"\x00" * params.k:
        raise SphinxException("Modified Body")

    delta = unpad_body(delta[params.k:])
    return decode(delta)

def receive_surb(params, keytuple, delta): 
    """Processes a SURB body to extract the reply. The keytuple was provided at the time of 
    SURB creation, and can be indexed by the SURB id, which is also returned to the receiving user.

    Returns the decoded message.
    """
    p = params
        
    ktilde = keytuple.pop(0)
    nu = len(keytuple)
    for i in range(nu-1, -1, -1):
        delta = p.pi(keytuple[i], delta)
    delta = p.pii(ktilde, delta)

    if delta[:p.k] == (b"\x00" * p.k):
        msg = unpad_body(delta[p.k:])
    
    return msg

def pack_message(params, m):
    """ A method to pack mix messages. """
    return encode(((params.max_len, params.m), m))

def unpack_message(params_dict, m):
    """ A method to unpack mix messages. """
    lens, msg = decode(m)
    if tuple(lens) not in params_dict:
        raise SphinxException("No parameter settings for: %s" % lens)
    return params_dict[tuple(lens)], msg

def test_timing():
    r = 5
    params = SphinxParams()
    pki = {}
    
    pkiPriv = {}
    pkiPub = {}

    for i in range(10):
        nid = pack("b", i)
        x = params.group.gensecret()
        y = params.group.expon(params.group.g, x)
        pkiPriv[nid] = pki_entry(nid, x, y)
        pkiPub[nid] = pki_entry(nid, None, y)

    # The simplest path selection algorithm and message packaging
    use_nodes = rand_subset(pkiPub.keys(), r)
    nodes_routing = list(map(Nenc, use_nodes))
    node_keys = [pkiPub[n].y for n in use_nodes]
    print()
    
    import time
    t0 = time.time()
    for _ in range(100):
        header, delta = create_forward_message(params, nodes_routing, node_keys, b"dest", b"this is a test")
    t1 = time.time()
    print("Time per mix encoding: %.2fms" % ((t1-t0)*1000.0/100))

    from .SphinxNode import sphinx_process
    import time
    t0 = time.time()
    for _ in range(100):
        x = pkiPriv[use_nodes[0]].x
        sphinx_process(params, x, header, delta)
    t1 = time.time()
    print("Time per mix processing: %.2fms" % ((t1-t0)*1000.0/100))


def test_minimal():
    from .SphinxParams import SphinxParams
    r = 5
    params = SphinxParams()

    # The minimal PKI involves names of nodes and keys
    
    pkiPriv = {}
    pkiPub = {}

    for i in range(10):
        nid = pack("b", i) # Nenc(params, bytes([i]))
        x = params.group.gensecret()
        y = params.group.expon(params.group.g, x)
        pkiPriv[nid] = pki_entry(nid, x, y)
        pkiPub[nid] = pki_entry(nid, None, y)

    # The simplest path selection algorithm and message packaging
    use_nodes = rand_subset(pkiPub.keys(), r)
    nodes_routing = list(map(Nenc, use_nodes))
    node_keys = [pkiPub[n].y for n in use_nodes]
    dest = b"bob"
    message = b"this is a test"
    header, delta = create_forward_message(params, nodes_routing, node_keys, dest, message)

    # Test encoding and decoding

    bin_message = pack_message(params, (header, delta))
    param_dict = { (params.max_len, params.m):params }

    px, (header1, delta1) = unpack_message(param_dict, bin_message)
    assert px == params
    assert header == tuple(header1)
    assert delta == delta1

    # Process message by the sequence of mixes
    from .SphinxNode import sphinx_process
    x = pkiPriv[use_nodes[0]].x

    i = 0
    while True:

        ret = sphinx_process(params, x, header, delta)
        (tag, B, (header, delta)) = ret
        routing = PFdecode(params, B)

        print("round %d" % i)
        i += 1

        # print("Type: %s" % typex)
        if routing[0] == Relay_flag:
            addr = routing[1]
            x = pkiPriv[addr].x 
        elif routing[0] == Dest_flag:
            assert len(routing) == 1
            assert delta[:16] == b"\x00" * params.k
            dec_dest, dec_msg = receive_forward(params, delta)
            assert dec_dest == dest
            assert dec_msg == message

            break
        else:
            print("Error")
            assert False
            break

    # Test the nym creation
    surbid, surbkeytuple, nymtuple = create_surb(params, nodes_routing, node_keys, b"myself")
    
    message = b"This is a reply"
    header, delta = package_surb(params, nymtuple, message)

    x = pkiPriv[use_nodes[0]].x

    while True:
        ret = sphinx_process(params, x, header, delta)
        (tag, B, (header, delta)) = ret
        routing = PFdecode(params, B)

        if routing[0] == Relay_flag:
            flag, addr = routing
            x = pkiPriv[addr].x 
        elif routing[0] == Surb_flag:
            flag, dest, myid = routing
            break

    received = receive_surb(params, surbkeytuple, delta)
    assert received == message

if __name__ == "__main__":
    test_timing() 


#!/usr/bin/env python

# Copyright 2011 Ian Goldberg
# Copyright 2016 George Danezis (UCL InfoSec Group)
#
# This file is part of Sphinx.
# 
# Sphinx is free software: you can redistribute it and/or modify
# it under the terms of version 3 of the GNU Lesser General Public
# License as published by the Free Software Foundation.
# 
# Sphinx is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public
# License along with Sphinx.  If not, see
# <http://www.gnu.org/licenses/>.
#
# The LIONESS implementation and the xcounter CTR mode class are adapted
# from "Experimental implementation of the sphinx cryptographic mix
# packet format by George Danezis".


from os import urandom
from hashlib import sha256
import hmac

from petlib.ec import EcGroup, EcPt, POINT_CONVERSION_UNCOMPRESSED
from petlib.bn import Bn
from petlib.cipher import Cipher

# Python 2/3 compatibility
from builtins import bytes

class Group_ECC:
    "Group operations in ECC"

    def __init__(self, gid=713):
        self.G = EcGroup(gid)
        self.g = self.G.generator()

    def gensecret(self):
        return self.G.order().random()

    def expon(self, base, exp):        
        x = exp
        b = base
        return (x * b)

    def multiexpon(self, base, exps):
        expon = 1
        for e in exps:
            expon = e.mod_mul( expon, self.G.order())
        return (expon * base)

    def makeexp(self, data):
        return (Bn.from_binary(data) % self.G.order())

    def in_group(self, alpha):
        # All strings of length 32 are in the group, says DJB
        b = alpha
        return self.G.check_point(b)

    def printable(self, alpha):
        return alpha.export(POINT_CONVERSION_UNCOMPRESSED)

def test_group():
    G = Group_ECC()
    sec1 = G.gensecret();
    sec2 = G.gensecret();
    gen = G.g

    assert G.expon(G.expon(gen, sec1), sec2) == G.expon(G.expon(gen, sec2), sec1)
    assert G.expon(G.expon(gen, sec1), sec2) == G.multiexpon(gen, [sec2, sec1])
    assert G.in_group(G.expon(gen, sec1))

def test_params():
    # Test Init
    params = SphinxParams()
    
    # Test Lioness
    k = b"A" * 16
    m = b"ARG"* 16

    c = params.lioness_enc(k,m)
    m2 = params.lioness_dec(k, c)
    assert m == m2

    k = urandom(16)
    c = params.aes_ctr(k, b"Hello World!")
    assert params.aes_ctr(k, c) == b"Hello World!"

class SphinxParams:

    def __init__(self, group=None, header_len = 192, body_len = 1024):
        # self.r = r
        self.aes = Cipher("AES-128-CTR")

        self.max_len = header_len
        self.m = body_len
        self.k = 16

        self.group = group
        if not group:
            self.group = Group_ECC()


    # The LIONESS PRP

    def aes_ctr(self, k, m, iv = b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"):
        k = bytes(k)
        m = bytes(m)
        assert type(k) is bytes and type(m) is bytes
        c = self.aes.enc(k, iv).update(m)
        return bytes(c)

    def lioness_enc(self, key, message):
        assert len(key) == self.k
        assert len(message) >= self.k * 2

        # Round 1
        k1 = self.hash(message[self.k:]+key+b'1')[:self.k]
        c = self.aes_ctr(key, message[:self.k], iv = k1)
        r1 = c + message[self.k:]

        # Round 2
        c = self.aes_ctr(key, r1[self.k:], iv = r1[:self.k])
        r2 = r1[:self.k] + c

        # Round 3
        k3 = self.hash(r2[self.k:]+key+b'3')[:self.k]
        c = self.aes_ctr(key, r2[:self.k], iv = k3)
        r3 = c + r2[self.k:]

        # Round 4
        c = self.aes_ctr(key, r3[self.k:], r3[:self.k])
        r4 = r3[:self.k] + c

        return r4

    def lioness_dec(self, key, message):
        assert len(key) == self.k
        assert len(message) >= self.k * 2

        r4 = message
        r4_short, r4_long = r4[:self.k], r4[self.k:]

        # Round 4        
        r3_long = self.aes_ctr(key, r4_long, iv = r4_short)
        r3_short = r4_short 

        # Round 3
        k2 = self.hash(r3_long+key+b'3')[:self.k]
        r2_short = self.aes_ctr(key, r3_short, iv = k2)
        r2_long = r3_long

        # Round 2
        r1_long = self.aes_ctr(key, r2_long, iv = r2_short)
        r1_short = r2_short 

        # Round 1
        k0 = self.hash(r1_long+key+b'1')[:self.k]
        c = self.aes_ctr(key, r1_short, iv = k0)
        r0 = c + r1_long

        return r0


    def xor_rho(self, key, plain):
        assert len(key) == self.k
        # p = b"\x00" * other
        return self.aes_ctr(key, plain)


    # The HMAC; key is of length k, output is of length k
    def mu(self, key, data):
        mac = hmac.new(key, data, digestmod=sha256).digest()[:self.k]
        return mac

    # The PRP; key is of length k, data is of length m
    def pi(self, key, data):
        assert len(key) == self.k
        assert len(data) == self.m

        return self.lioness_enc(key, data)

    # The inverse PRP; key is of length k, data is of length m
    def pii(self, key, data):
        assert len(key) == self.k
        assert len(data) == self.m

        return self.lioness_dec(key, data)

    # The various hashes

    def hash(self, data):
        return sha256(data).digest()

    def get_aes_key(self, s):
        group = self.group
        return self.hash(b"aes_key:" + group.printable(s))[:self.k]

    def derive_key(self, k, flavor):
        iv = flavor
        m = b"\x00" * self.k
        K = self.aes.enc(k, iv).update(m)
        return K

    def hb(self, alpha, k):
        "Compute a hash of alpha and s to use as a blinding factor"
        K = self.derive_key(k, b"hbhbhbhbhbhbhbhb")
        return self.group.makeexp(K)

    def hrho(self, k):
        "Compute a hash of s to use as a key for the PRG rho"
        K = self.derive_key(k, b"hrhohrhohrhohrho")
        return K

    def hmu(self, k):
        "Compute a hash of s to use as a key for the HMAC mu"
        K = self.derive_key(k, b"hmu:hmu:hmu:hmu:")
        return K

    def hpi(self, k):
        "Compute a hash of s to use as a key for the PRP pi"
        K = self.derive_key(k, b"hpi:hpi:hpi:hpi:")
        return K

    def htau(self, k):
        "Compute a hash of s to use to see if we've seen s before"
        K = self.derive_key(k, b"htauhtauhtauhtau")
        return K


from setuptools import setup

import sphinxmix

if __name__ == "__main__":
      
      setup(name='sphinxmix',
            version=sphinxmix.VERSION,
            description='A Python implementation of the Sphinx mix packet format.',
            author='George Danezis',
            author_email='g.danezis@ucl.ac.uk',
            url=r'http://sphinxmix.readthedocs.io/en/latest/',
            packages=['sphinxmix'],
            license="2-clause BSD",
            long_description="""A Python implementation of the Sphinx mix packet format.

            For full documentation see: http://sphinxmix.readthedocs.io/en/latest/
            """,

            setup_requires=['pytest-runner', "pytest"],
            tests_require=[
                  "pytest",
                  "future >= 0.14.3",
                  "pytest >= 3.0.0",
                  "msgpack-python >= 0.4.6",
                  "petlib >= 0.0.38",
            ],
            install_requires=[
                  "future >= 0.14.3",
                  "pytest >= 3.0.0",
                  "msgpack-python >= 0.4.6",
                  "petlib >= 0.0.38",
            ]
      )